import math

import pyro
import pyro.distributions as dist
import torch
import torch.nn as nn
import torch.nn.functional as F
from pyro import poutine
from pyro.infer import Predictive
from torch.distributions.utils import broadcast_all

EPS = 1e-8
scale_init = math.log(0.01)


class MLPEncoderDirichlet(pyro.nn.PyroModule):
    def __init__(
        self,
        n_genes,
        hidden_size,
        n_topics,
        dropout,
        n_layers,
        num_nodes,
        n_batches,
    ):
        super().__init__()

        self.drop = nn.Dropout(dropout)
        self.bn_pp = nn.BatchNorm1d(n_genes * (n_layers + 1))

        self.linear = nn.Linear(n_genes * (n_layers + 1), hidden_size)
        self.mu_topic = nn.Linear(hidden_size, n_topics)
        self.norm_topic = nn.BatchNorm1d(
            n_topics, affine=False, track_running_stats=False
        )

        self.reset_parameters()

    def forward(self, x):
        x = self.drop(x)
        x = self.bn_pp(x)
        x = F.relu(self.linear(x))

        concent = self.mu_topic(x)
        concent = self.norm_topic(concent)
        concent = torch.clamp(concent, max=8)
        concent = concent.exp()
        return concent

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.linear.weight)
        nn.init.xavier_uniform_(self.mu_topic.weight)


class MLPEncoderMVN(pyro.nn.PyroModule):
    def __init__(
        self,
        n_genes,
        hidden_size,
        n_topics,
        dropout,
        n_layers,
        num_nodes,
        n_batches,
    ):
        super().__init__()

        self.n_topics = n_topics
        self.hidden_size = hidden_size

        self.drop = nn.Dropout(dropout)
        self.bn_pp = nn.BatchNorm1d(n_genes * (n_layers + 1))
        self.bn = nn.BatchNorm1d(hidden_size)
        self.norm_topic = nn.BatchNorm1d(
            n_topics, affine=False, track_running_stats=False
        )

        self.mu_topic = nn.Linear(hidden_size, n_topics)
        self.diag_topic = nn.Linear(hidden_size, n_topics)
        self.k = int(self.n_topics * (self.n_topics - 1) / 2)

        self.cov_factor = nn.Linear(hidden_size, self.k)
        self.linear = nn.Linear(n_genes * (n_layers + 1), hidden_size)

        self.reset_parameters()

    def forward(self, x):
        x = self.drop(x)
        x = self.bn_pp(x)
        x = self.linear(x)
        x = F.relu(self.bn(x))
        # x = self.drop(x)
        # x = F.relu(x)

        mu_topic = self.mu_topic(x)
        mu_topic = self.norm_topic(mu_topic)

        diag_topic = self.diag_topic(x)
        diag_topic = torch.sqrt(diag_topic.exp())
        cov_factor = self.cov_factor(x)
        indices = torch.tril_indices(row=self.n_topics, col=self.n_topics, offset=-1)

        cov = torch.zeros((x.shape[0], self.n_topics, self.n_topics), device=x.device)
        cov[:, indices[0], indices[1]] = cov_factor
        cov = cov + torch.diag_embed(diag_topic)

        return mu_topic, cov

    def reset_parameters(self):
        nn.init.xavier_normal_(self.linear.weight)
        nn.init.xavier_normal_(self.mu_topic.weight)
        nn.init.zeros_(self.diag_topic.weight)
        nn.init.zeros_(self.cov_factor.weight)
        nn.init.zeros_(self.cov_factor.bias)


# class BatchEncoder(pyro.nn.PyroModule):
#     def __init__(self, n_genes, n_layers, n_batches):
#         super().__init__()

#         if n_batches == 1:
#             n_batches = 0

#         self.bn_pp = nn.BatchNorm1d(n_genes * (n_layers + 1) + n_batches)
#         self.linear1 = nn.Linear(n_genes * (n_layers + 1) + n_batches, 20)
#         self.linear2 = nn.Linear(20, n_genes)
#         self.reset_parameters()

#     def forward(self, x):
#         x = torch.log(x + 1)
#         # x = self.bn_pp(x)
#         # mu = self.linear1(x)
#         # x = F.gelu(self.linear(x))
#         # mu = self.linear1(x)
#         x = self.linear1(x)
#         sigma = self.linear2(x)
#         # sigma = self.bn1(sigma)
#         sigma = F.softplus(sigma)  # (mu - mu.mean())/mu.std()
#         # return (mu - mu.mean()) / mu.std(), sigma
#         return sigma

#     def reset_parameters(self):
#         nn.init.xavier_uniform_(self.linear1.weight)
#         nn.init.xavier_uniform_(self.linear2.weight)


class spatialLDAModel(pyro.nn.PyroModule):
    def __init__(
        self,
        n_genes,
        hidden_size,
        n_topics,
        dropout,
        init_bg,
        n_layers,
        n_batches,
        num_nodes,
        enc_distribution,
        beta,
    ):
        super().__init__()

        self.n_topics = n_topics
        self.n_genes = n_genes
        self.n_batches = n_batches
        self.num_nodes = num_nodes
        self.n_topics = n_topics
        self.beta = beta
        self.enc_distribution = enc_distribution

        if self.enc_distribution == "mvn":
            self.encoder = MLPEncoderMVN(
                n_genes,
                hidden_size,
                n_topics,
                dropout,
                n_layers,
                num_nodes,
                n_batches,
            )
        else:
            self.encoder = MLPEncoderDirichlet(
                n_genes,
                hidden_size,
                n_topics,
                dropout,
                n_layers,
                num_nodes,
                n_batches,
            )

        # if self.mode == "cell-batch":
        # self.batchencoder = BatchEncoder(n_genes, n_layers, n_batches)
        self.register_buffer("init_bg", init_bg, persistent=False)
        self.register_buffer("alpha", torch.tensor(1 / self.n_topics), persistent=False)

    def model(self, x, sgc_x, ys=None):
        pyro.module("stamp", self)

        batch_size = x.shape[0]
        ls = torch.sum(x, -1, keepdim=True)

        # Global variables
        with poutine.scale(scale=batch_size / self.num_nodes):
            disp_aux = pyro.sample(
                "disp_aux",
                dist.HalfNormal(torch.ones(1, device=x.device) * 5),
            )

            with pyro.plate("covariates", self.n_batches):
                d_aux = pyro.sample(
                    "d_aux",
                    dist.HalfNormal(
                        torch.ones(self.n_batches, device=x.device) * 1
                    ).to_event(0),
                )

            with pyro.plate("genes", self.n_genes):
                disp = pyro.sample(
                    "disp",
                    dist.HalfCauchy(
                        torch.ones(1, device=x.device) * disp_aux,
                    ),
                )

                delta = pyro.sample(
                    "delta",
                    dist.HalfCauchy(torch.ones(self.n_genes, device=x.device)),
                )
                delta = delta**2
                delta = delta.unsqueeze(0).repeat([self.n_topics, 1])

                d = pyro.sample(
                    "d",
                    dist.Normal(
                        x.new_zeros(self.n_genes, self.n_batches),
                        x.new_ones(self.n_genes, self.n_batches) * d_aux,
                    ).to_event(1),
                )
                # ys_scale = torch.exp(ys_scale)
                d = d.t()

            with pyro.plate("topics", self.n_topics):
                tau = pyro.sample(
                    "tau",
                    dist.HalfCauchy(torch.ones(self.n_topics, device=x.device)),
                )
                tau = tau**2
                tau = tau.unsqueeze(-1).expand(self.n_topics, self.n_genes)

                # caux is the squared version
                caux = pyro.sample(
                    "caux",
                    dist.InverseGamma(
                        torch.ones(
                            (self.n_topics, self.n_genes),
                            device=x.device,
                        ),
                        torch.ones(
                            (self.n_topics, self.n_genes),
                            device=x.device,
                        ),
                    ).to_event(1),
                )
                # caux = caux.unsqueeze(-1).expand(self.n_topics, self.n_genes)
                # lambda_tilde is also the squared version
                lambda_ = pyro.sample(
                    "lambda_",
                    dist.HalfCauchy(
                        torch.ones((self.n_topics, self.n_genes), device=x.device)
                    ).to_event(1),
                )
                lambda_ = lambda_**2

                lambda_tilde = (caux * tau * delta * lambda_) / (
                    caux + tau * delta * lambda_
                )

                w = pyro.sample(
                    "w",
                    dist.Normal(
                        torch.zeros(
                            (self.n_topics, self.n_genes),
                            device=x.device,
                        ),
                        torch.ones(
                            (self.n_topics, self.n_genes),
                            device=x.device,
                        )
                        * torch.sqrt(lambda_tilde),
                    ).to_event(1),
                )

        with pyro.plate("batch", batch_size):
            s = pyro.param("s", torch.zeros(1, device=x.device))
            s = torch.sqrt(s.exp())

            if self.enc_distribution == "mvn":
                z_topic_loc = x.new_zeros((batch_size, self.n_topics))
                z_topic_scale = x.new_ones((batch_size, self.n_topics)) * s
                with poutine.scale(scale=self.beta):
                    z_topic = pyro.sample(
                        "z_topic",
                        dist.Normal(z_topic_loc, z_topic_scale).to_event(1),
                    )
                    z = torch.exp(F.log_softmax(z_topic, dim=1))
            else:
                with poutine.scale(scale=self.beta):
                    z_topic = pyro.sample(
                        "z_topic",
                        dist.Dirichlet(
                            self.alpha.repeat(batch_size, self.n_topics)
                        ).to_event(0),
                    )
                    z = z_topic

            if ys is not None:
                d = ys @ d

            mean = torch.exp(
                F.log_softmax(
                    d + torch.log(z @ F.softmax(w + self.init_bg, dim=1)), dim=1
                )
            )

            rate = 1 / disp
            mean, rate = broadcast_all(mean, rate)
            mean = ls * mean
            pyro.sample("obs", dist.GammaPoisson(rate, rate / mean).to_event(1), obs=x)

    def guide(self, x, sgc_x=None, ys=None):
        pyro.module("stamp", self)
        batch_size = x.shape[0]

        with poutine.scale(scale=batch_size / self.num_nodes):
            disp_aux_loc = pyro.param(
                "disp_aux_loc",
                torch.zeros(1, device=x.device),
            )
            disp_aux_scale = pyro.param(
                "disp_aux_scale",
                torch.ones(1, device=x.device) * scale_init,
            )
            disp_aux_scale = torch.sqrt(torch.exp(disp_aux_scale))
            pyro.sample("disp_aux", dist.LogNormal(disp_aux_loc, disp_aux_scale))

            with pyro.plate("covariates", self.n_batches):
                d_aux_loc = pyro.param(
                    "d_aux_loc",
                    torch.zeros(self.n_batches, device=x.device),
                )
                d_aux_scale = pyro.param(
                    "d_aux_scale",
                    torch.ones(self.n_batches, device=x.device) * scale_init,
                )
                d_aux_scale = torch.sqrt(torch.exp(d_aux_scale))
                pyro.sample("d_aux", dist.LogNormal(d_aux_loc, d_aux_scale).to_event(0))

            with pyro.plate("genes", self.n_genes):
                disp_loc = pyro.param(
                    "disp_loc",
                    torch.zeros(self.n_genes, device=x.device),
                )
                disp_scale = pyro.param(
                    "disp_scale",
                    torch.ones(self.n_genes, device=x.device) * scale_init,
                )
                disp_scale = torch.sqrt(torch.exp(disp_scale))
                pyro.sample("disp", dist.LogNormal(disp_loc, disp_scale))

                delta_loc = pyro.param(
                    "delta_loc",
                    torch.zeros((self.n_genes), device=x.device),
                )
                delta_scale = pyro.param(
                    "delta_scale",
                    torch.ones((self.n_genes), device=x.device) * scale_init,
                )
                delta_scale = torch.sqrt(torch.exp(delta_scale))
                pyro.sample("delta", dist.LogNormal(delta_loc, delta_scale))

                d_loc = pyro.param(
                    "d_loc",
                    torch.zeros((self.n_genes, self.n_batches), device=x.device),
                )
                d_scale = pyro.param(
                    "d_scale",
                    torch.ones((self.n_genes, self.n_batches), device=x.device)
                    * scale_init,
                )
                d_scale = torch.sqrt(torch.exp(d_scale))
                pyro.sample(
                    "d",
                    dist.Normal(
                        d_loc,
                        d_scale,
                    ).to_event(1),
                )

            with pyro.plate("topics", self.n_topics):
                tau_loc = pyro.param(
                    "tau_loc",
                    torch.zeros((self.n_topics), device=x.device),
                )
                tau_scale = pyro.param(
                    "tau_scale",
                    torch.ones((self.n_topics), device=x.device) * scale_init,
                )
                tau_scale = torch.sqrt(torch.exp(tau_scale))
                pyro.sample("tau", dist.LogNormal(tau_loc, tau_scale))

                caux_loc = pyro.param(
                    "caux_loc",
                    torch.zeros((self.n_topics, self.n_genes), device=x.device),
                )
                caux_scale = pyro.param(
                    "caux_scale",
                    torch.ones((self.n_topics, self.n_genes), device=x.device)
                    * scale_init,
                )
                caux_scale = torch.sqrt(torch.exp(caux_scale))
                pyro.sample("caux", dist.LogNormal(caux_loc, caux_scale).to_event(1))

                lambda_loc = pyro.param(
                    "lambda_loc",
                    torch.zeros((self.n_topics, self.n_genes), device=x.device),
                )
                lambda_scale = pyro.param(
                    "lambda_scale",
                    torch.ones((self.n_topics, self.n_genes), device=x.device)
                    * scale_init,
                )
                lambda_scale = torch.sqrt(torch.exp(lambda_scale))
                pyro.sample(
                    "lambda_", dist.LogNormal(lambda_loc, lambda_scale).to_event(1)
                )

                w_loc = pyro.param(
                    "w_loc",
                    torch.zeros((self.n_topics, self.n_genes), device=x.device),
                )
                w_scale = pyro.param(
                    "w_scale",
                    torch.ones((self.n_topics, self.n_genes), device=x.device)
                    * scale_init,
                )
                w_scale = torch.sqrt(torch.exp(w_scale))
                pyro.sample("w", dist.Normal(w_loc, w_scale).to_event(1))

        with pyro.plate("batch", batch_size):
            if self.enc_distribution == "mvn":
                z_loc, z_cov = self.encoder(sgc_x)
                with poutine.scale(scale=self.beta):
                    pyro.sample(
                        "z_topic",
                        dist.MultivariateNormal(z_loc, scale_tril=z_cov).to_event(0),
                    )
                z_loc = F.softmax(z_loc, dim=1)
            else:
                z_loc = self.encoder(sgc_x)

                with poutine.scale(scale=self.beta):
                    pyro.sample("z_topic", dist.Dirichlet(z_loc).to_event(0))
                z_loc = dist.Dirichlet(z_loc).mean

            return z_loc

    def feature_by_topic(self, return_scale, return_softmax=True):
        if return_scale:
            return pyro.param("w_loc").t(), pyro.param("w_scale").t()
        w = pyro.param("w_loc")
        if return_softmax:
            w = F.softmax(w + self.init_bg, dim=1)
        return w.t()

    def predictive(self, num_samples):
        return Predictive(self.model, guide=self.guide, num_samples=num_samples)

    def feature(self, param):
        return pyro.param(param)

    def get_bias(self):
        return self.init_bg

    def model_params(self):
        return pyro.module("stamp", self)

    def set_device(self, device):
        param_state = pyro.get_param_store().get_state()
        with torch.no_grad():
            for name, value in param_state["params"].items():
                param_state["params"][name] = value.to(device)

        pyro.get_param_store().set_state(param_state)



"""Top-level package for scTM."""

__author__ = """Zhong Chengwei"""
__email__ = "chengwei8@gmail.com"
__version__ = "0.1.0"

from . import pl, pp, stamp, utils, analysis, metrics

__all__ = ["pl", "pp", "analysis", "stamp", "utils", "metrics"]



import numpy as np
import itertools
from scipy.sparse import issparse
import pandas as pd
from scipy.stats import spearmanr


def check_layer(adata, layer):
    if layer is None:
        if issparse(adata.X):
            data = adata.X.toarray()
        else:
            data = adata.X
    else:
        if issparse(adata.layers[layer]):
            data = adata.layers[layer].toarray()
        else:
            data = adata.layers[layer]
    return data


def get_topic_diversity(beta, topk=10):
    num_topics = beta.shape[1]
    topics = [beta.nlargest(topk, i).index.tolist() for i in beta.columns]
    genes = list(itertools.chain(*topics))
    n_unique = len(np.unique(genes))
    TD = n_unique / (topk * num_topics)
    return TD


def get_cell_probability(data, wi, quantiles_df, wj=None):

    if wj is None:
        D_wi = (data[wi] >= np.max((quantiles_df.loc[wi, "quantiles"], 1))).mean()
        return D_wi

    # Find probability that they are not both zero
    D_wj = (data[wj] >= np.max((quantiles_df.loc[wj, "quantiles"], 1))).mean()
    D_wi_wj = (
        (data[wi] >= np.max((quantiles_df.loc[wi, "quantiles"], 1)))
        & (data[wj] >= np.max((quantiles_df.loc[wj, "quantiles"], 1)))
    ).mean()

    return D_wj, D_wi_wj


def get_topic_coherence(
    adata, beta, layer=None, topk=10, quantile=0.8, individual=False
):

    data = check_layer(adata, layer)
    data = data.copy()
    data = pd.DataFrame(data, index=adata.obs_names, columns=adata.var_names)
    quantiles = np.quantile(data, q=quantile, axis=0)
    quantiles_df = pd.DataFrame(quantiles, index=adata.var_names, columns=["quantiles"])

    TC = []
    topics = [beta.nlargest(topk, i).index.tolist() for i in beta.columns]
    for beta_topk in topics:
        TC_k = 0
        counter = 0
        for i, gene in enumerate(beta_topk):
            # get D(w_i)
            D_wi = get_cell_probability(data, gene, quantiles_df)
            j = i + 1
            tmp = 0
            while j < len(beta_topk) and j > i:
                # get D(w_j) and D(w_i, w_j)
                D_wj, D_wi_wj = get_cell_probability(
                    data, gene, quantiles_df, beta_topk[j]
                )
                # get f(w_i, w_j)
                if D_wi_wj == 0:
                    f_wi_wj = 0
                else:
                    # f_wi_wj = -1 + (np.log(D_wi) + np.log(D_wj) - 2.0 * np.log(D)) / (
                    #     np.log(D_wi_wj) - np.log(D)
                    # )
                    f_wi_wj = (np.log(D_wi_wj) - np.log(D_wi) - np.log(D_wj)) / (
                        -np.log(D_wi_wj + 1e-5)
                    )
                    # f_wi_wj = (np.log(D_wi) + np.log(D_wj)) / (np.log(D_wi_wj)) -1
                # update tmp:
                tmp += f_wi_wj
                j += 1
                counter += 1
            # update TC_k
            TC_k += tmp
        TC.append(TC_k / counter)

    if individual:
        return TC
    else:
        TC = np.mean(TC)
        return TC


def get_gene_topic_coherence(
    adata, beta, topic_prop, topk=10, layer=None, individual=False
):

    data = check_layer(adata, layer)

    data = pd.DataFrame(data, index=adata.obs_names, columns=adata.var_names)
    # Drop batch
    topics = [beta.nlargest(topk, i).index.tolist() for i in beta.columns]
    TGC = []
    for i in range(len(topics)):
        topic_genes = topics[i]
        TGC_topic = []
        for topic_gene in topic_genes:
            cor = spearmanr(data[topic_gene], topic_prop.iloc[:, i])[0]
            TGC_topic.append(cor)
        TGC.append(np.mean(TGC_topic))

    if not individual:
        return np.mean(TGC)
    else:
        return TGC


def get_metrics(adata, beta, topic_prop, topk=10, layer=None, TGC=True):

    TC = get_topic_coherence(adata, beta, layer=layer, topk=topk)
    TD = get_topic_diversity(beta, topk=topk)
    if TGC:
        TGC = get_gene_topic_coherence(adata, beta, topic_prop, topk, layer)
        return {
            "Topic Coherence": TC,
            "Topic Diversity": TD,
            "Gene Topic Coherence": TGC,
        }
    else:
        return {"Topic Coherence": TC, "Topic Diversity": TD}



import anndata as ad
import scanpy as sc
import matplotlib.pyplot as plt
import textwrap
import numpy as np
from matplotlib.axes import Axes
import pandas as pd
from matplotlib.patches import Patch
import seaborn as sns
from matplotlib import rcParams
import matplotlib as mpl

# import matplotlib.pyplot as plt
# import numpy as np
from matplotlib.colors import ListedColormap
from matplotlib.gridspec import GridSpec

# from upsetplot import plot, from_contents
from itertools import chain


from scanpy._utils import _empty, Empty
from scanpy.pl._tools.scatterplots import (
    _check_spatial_data,
    _check_img,
    _check_spot_size,
    _check_scale_factor,
    _check_crop_coord,
    _check_na_color,
)
from typing import (
    Union,
    Optional,
    List,
    Tuple,
)
from scanpy.pl._utils import (
    ColorLike,
)


def heatmap_topic(adata, groupby=None, topics=None, figsize=(10, 5), cmap=None):
    topic_prop = adata.obs.copy()

    if topics is None:
        topics = topic_prop.columns[topic_prop.columns.str.startswith("Topic")]
        topic_prop = topic_prop.loc[:, topics]
    else:
        topic_prop = topic_prop.loc[:, topics]

    topic_adata = ad.AnnData(topic_prop)
    sc.pl.clustermap(topic_adata, obs_key=groupby, figsize=figsize, cmap=cmap)


def heatmap(
    adata,
    topic_prop,
    groupby=None,
    figsize=(10, 5),
    dendrogram=False,
    swap_axes=True,
    cmap=None,
    save=None,
    **kwargs,
):
    topic_adata = ad.AnnData(topic_prop)
    topic_adata.obs[groupby] = adata.obs[groupby]

    sc.pl.heatmap(
        adata=topic_adata,
        var_names=topic_prop.columns,
        groupby=groupby,
        figsize=figsize,
        dendrogram=dendrogram,
        swap_axes=swap_axes,
        cmap=cmap,
        save=save,
        **kwargs,
    )


def matrixplot(
    adata,
    groupby=None,
    topics=None,
    figsize=(10, 5),
    dendrogram=False,
    swap_axes=True,
    cmap=None,
    save=None,
):
    topic_prop = adata.obs.copy()

    if topics is None:
        topics = topic_prop.columns[topic_prop.columns.str.startswith("Topic")]
        topic_prop = topic_prop.loc[:, topics]
    else:
        topic_prop = topic_prop.loc[:, topics]

    topic_adata = ad.AnnData(topic_prop)
    topic_adata.obs[groupby] = adata.obs[groupby]
    sc.pl.matrixplot(
        adata=topic_adata,
        var_names=topics,
        groupby=groupby,
        figsize=figsize,
        dendrogram=dendrogram,
        swap_axes=swap_axes,
        cmap=cmap,
        save=save,
    )


def tracksplot(
    adata,
    groupby=None,
    topics=None,
    figsize=(10, 5),
    dendrogram=False,
    swap_axes=True,
    cmap=None,
    save=None,
):
    topic_prop = adata.obs.copy()

    if topics is None:
        topics = topic_prop.columns[topic_prop.columns.str.startswith("Topic")]
        topic_prop = topic_prop.loc[:, topics]
    else:
        topic_prop = topic_prop.loc[:, topics]

    topic_adata = ad.AnnData(topic_prop)
    topic_adata.obs[groupby] = adata.obs[groupby]
    sc.pl.tracksplot(
        adata=topic_adata,
        var_names=topics,
        groupby=groupby,
        figsize=figsize,
        dendrogram=dendrogram,
        swap_axes=swap_axes,
        cmap=cmap,
        save=save,
    )


def clustermap(
    beta,
    gene_dict=None,
    topn_genes=20,
    xticklabels="auto",
    yticklabels="auto",
    figsize=(10, 6),
    cmap="viridis",
    fontsize=5,
    row_cluster=False,
    col_cluster=True,
    standard_scale=0,
    transpose=False,
    return_fig=False,
):
    if gene_dict is None:
        genes = []
        topics = beta.columns
        for i in topics:
            genes.append(beta.nlargest(topn_genes, i).index.tolist())
        genes = list(set(list(chain.from_iterable(genes))))
        beta_sub = beta.loc[genes, :]
    else:
        genes = [x for x in gene_dict.values()]
        genes = list(chain.from_iterable(genes))
        beta_sub = beta.loc[genes, :]

    if transpose:
        beta_sub = beta_sub.transpose()

    fig = sns.clustermap(
        beta_sub,
        cmap=cmap,
        figsize=figsize,
        row_cluster=row_cluster,
        col_cluster=col_cluster,
        standard_scale=standard_scale,
        xticklabels=xticklabels,
        yticklabels=yticklabels,
    )

    fig.fig.subplots_adjust(right=0.7)
    fig.ax_cbar.set_position((0.8, 0.4, 0.01, 0.3))

    fig.ax_heatmap.set_yticklabels(
        fig.ax_heatmap.get_ymajorticklabels(), fontsize=fontsize, rotation=0
    )

    # fig.ax_row_dendrogram.set_visible(False) #suppress row dendrogram
    # fig.ax_col_dendrogram.set_visible(False) #suppress column dendrogram

    if return_fig:
        fig = fig.fig
        return fig


def heatmap_topic_correlation(
    topic_prop,
    spatial_connectivities=None,
    return_values=False,
    figsize=(8, 6),
    cmap="viridis",
    fontsize=8,
):
    fig, ax = plt.subplots(figsize=figsize)

    if spatial_connectivities is None:
        corr = topic_prop.corr()
        sns.heatmap(
            corr,
            annot=True,
            vmin=-1,
            vmax=1,
            cmap=cmap,
            ax=ax,
            annot_kws={"fontsize": fontsize},
            fmt=".2f",
        )
    else:
        spatial_topic_prop = spatial_connectivities @ topic_prop
        spatial_topic_prop = pd.DataFrame(
            spatial_topic_prop, index=topic_prop.index, columns=topic_prop.columns
        )
        corr = spatial_topic_prop.corr()
        sns.heatmap(
            corr,
            annot=True,
            vmin=-1,
            vmax=1,
            cmap=cmap,
            ax=ax,
            annot_kws={"fontsize": fontsize},
            fmt=".2f",
        )
    if return_values:
        return corr
    else:
        return ax


def enrichment_barplot(
    enrichments,
    topic,
    type="enrichr",
    figsize=(10, 5),
    n_enrichments=5,
    qval_cutoff=0.05,
    title="auto",
):
    if type == "enrichr":
        if title == "auto":
            title = enrichments[topic]["Gene_set"][0]
        enrichment = enrichments[topic]
        enrichment = enrichment.loc[enrichment["Adjusted P-value"] < qval_cutoff, :]
        enrichment = enrichment.sort_values("Adjusted P-value")
        enrichment = enrichment.iloc[:n_enrichments, :]

        fig, ax = plt.subplots(figsize=figsize)
        ax.barh(
            y=enrichment["Term"],
            width=-np.log(enrichment["Adjusted P-value"]),
            fill="blue",
            align="center",
        )

        ax.set_yticklabels(
            [textwrap.fill(term, 24) for term in enrichment["Term"].values]
        )

        ax.set_xlabel("- Log Adjusted P-value")
        ax.set_title(title)

        ax.invert_yaxis()

        plt.tight_layout()
        return ax

    if type == "gsea":
        if title == "auto":
            title = enrichments[topic]["Name"][0]

        enrichment = enrichments[topic]
        enrichment = enrichment.loc[enrichment["NOM p-val"] < qval_cutoff, :]
        enrichment = enrichment[enrichment["NES"] > 0]
        enrichment = enrichment.sort_values("NES", ascending=False)
        enrichment["Term"] = enrichment["Term"].str.replace("_", " ")
        enrichment = enrichment.iloc[:n_enrichments, :]

        enrichment["-log_qval"] = -np.log(
            enrichment["FDR q-val"].astype("float") + 1e-7
        )

        fig, ax = plt.subplots(figsize=figsize)

        ax.barh(y=enrichment["Term"], width=enrichment["NES"], align="center")

        ax.set_xlabel("NES")
        ax.set_title(title)

        ax.set_yticklabels(
            [textwrap.fill(term, 24) for term in enrichment["Term"].values]
        )

        ax.invert_yaxis()
        plt.tight_layout()

        return ax


def enrichment_dotplot(
    enrichment,
    type="enrichr",
    figsize=(10, 5),
    n_enrichments=10,
    title="auto",
    cmap=None,
):
    fig, ax = plt.subplots(figsize=figsize)

    if type == "enrichr":
        # enrichment = enrichments[topic].copy()
        # enrichment = enrichment.loc[enrichment["Adjusted P-value"] < qval_cutoff, :]

        enrichment["gene_size"] = enrichment["Overlap"].str.split("/").str[1]
        enrichment["-log_qval"] = -np.log(enrichment["Adjusted P-value"])
        enrichment["gene_ratio"] = enrichment["Overlap"].str.split("/").str[0].astype(
            "int"
        ) / enrichment["Overlap"].str.split("/").str[1].astype("int")

        if enrichment.shape[0] < n_enrichments:
            n_enrichments = enrichment.shape[0]

        enrichment = enrichment.sort_values("gene_ratio")
        enrichment = enrichment.iloc[:n_enrichments, :]

        scatter = ax.scatter(
            x=enrichment["gene_ratio"].values,
            y=enrichment["Term"].values,
            s=enrichment["gene_size"].values.astype("int"),
            c=enrichment["Combined Score"].values,
            cmap=cmap,
        )
        ax.set_xlabel("Gene Ratio")

        legend1 = ax.legend(
            *scatter.legend_elements(prop="sizes", num=5),
            bbox_to_anchor=(1.04, 1),
            loc="upper left",
            title="Geneset Size",
            labelspacing=1,
            borderpad=1,
        )
        ax.legend(
            *scatter.legend_elements(prop="colors", num=5),
            bbox_to_anchor=(1.04, 0),
            loc="lower left",
            title="Combined Score",
            labelspacing=1,
            borderpad=1,
        )

        ax.add_artist(legend1)
        # ax.add_artist(legend2)

        ax.set_yticklabels(
            [textwrap.fill(term, 24) for term in enrichment["Term"].values]
        )

        if title == "auto":
            ax.set_title(enrichment["Gene_set"].values[0])

    if type == "gsea":
        # enrichment = enrichments[topic].copy()
        # enrichment = enrichment.loc[enrichment["FDR q-val"] < qval_cutoff, :]

        enrichment["gene_size"] = enrichment["Tag %"].str.split("/").str[1]
        enrichment["-log_qval"] = -np.log(
            enrichment["FDR q-val"].astype("float") + 1e-7
        )
        enrichment["gene_ratio"] = enrichment["Tag %"].str.split("/").str[0].astype(
            "int"
        ) / enrichment["Tag %"].str.split("/").str[1].astype("int")

        if enrichment.shape[0] < n_enrichments:
            n_enrichments = enrichment.shape[0]

        enrichment = enrichment.sort_values("-log_qval", ascending=False)
        enrichment = enrichment.iloc[:n_enrichments, :]

        scatter = ax.scatter(
            x=enrichment["-log_qval"].values,
            y=enrichment["Term"].values,
            s=enrichment["gene_ratio"].values.astype("float"),
            c=enrichment["NES"].values,
            cmap=cmap,
        )
        ax.set_xlabel("-log q_val")

        legend1 = ax.legend(
            *scatter.legend_elements(prop="sizes", num=5),
            bbox_to_anchor=(1, 1),
            loc="upper left",
            title="Gene Ratio",
            labelspacing=1,
            borderpad=1,
        )

        ax.legend(
            *scatter.legend_elements(prop="colors", num=5),
            bbox_to_anchor=(1, 0),
            loc="lower left",
            title="NES",
            labelspacing=1,
            borderpad=1,
        )

        ax.add_artist(legend1)
        # ax.add_artist(legend2)

        ax.set_yticklabels(
            [textwrap.fill(term, 30) for term in enrichment["Term"].values]
        )

        if title == "auto":
            ax.set_title(enrichment["Name"].values[0])

        ax.invert_yaxis()
        plt.tight_layout()

        return ax


def draw_pie(dist, xpos, ypos, size, colors, figsize, ax=None):
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    # for incremental pie slices
    cumsum = np.cumsum(dist)
    cumsum = cumsum / cumsum[-1]
    pie = [0] + cumsum.tolist()
    c = 0
    for r1, r2 in zip(pie[:-1], pie[1:]):
        # if r2 - r1 > 0.01:
        angles = np.linspace(2 * np.pi * r1, 2 * np.pi * r2, num=10)
        x = [0] + np.cos(angles).tolist() + [0]
        y = [0] + np.sin(angles).tolist() + [0]

        xy = np.column_stack([x, y])
        # print(xy.shape)
        ax.plot(
            [xpos],
            [ypos],
            marker=xy,
            markersize=np.abs(xy).max() * np.array(np.sqrt(size)),
            c=colors[c],
        )
        c += 1

    return ax


def spatialpie(
    adata,
    topic_prop,
    *,
    basis: str = "spatial",
    img: Union[np.ndarray, None] = None,
    img_key: Union[str, None, Empty] = _empty,
    library_id: Union[str, Empty] = _empty,
    crop_coord: Tuple[int, int, int, int] = None,
    alpha_img: float = 1.0,
    bw: Optional[bool] = False,
    frameon=False,
    size: float = 1.0,
    scale_factor: Optional[float] = None,
    spot_size: Optional[float] = None,
    na_color: Optional[ColorLike] = None,
    show: Optional[bool] = None,
    return_fig: Optional[bool] = None,
    cmap="tab20",
    legend=True,
    figsize=None,
    title=None,
    **kwargs,
) -> Union[Axes, List[Axes], None]:
    """
    Scatter plot in spatial coordinates.
    This function allows overlaying data on top of images.
    Use the parameter `img_key` to see the image in the background
    And the parameter `library_id` to select the image.
    By default, `'hires'` and `'lowres'` are attempted.
    Use `crop_coord`, `alpha_img`, and `bw` to control how it is displayed.
    Use `size` to scale the size of the Visium spots plotted on top.
    As this function is designed to for imaging data, there are two key assumptions
    about how coordinates are handled:
    1. The origin (e.g `(0, 0)`) is at the top left – as is common convention
    with image data.
    2. Coordinates are in the pixel space of the source image, so an equal
    aspect ratio is assumed.
    If your anndata object has a `"spatial"` entry in `.uns`, the `img_key`
    and `library_id` parameters to find values for `img`, `scale_factor`,
    and `spot_size` arguments. Alternatively, these values be passed directly.
    Parameters
    ----------
    {adata_color_etc}
    {scatter_spatial}
    {scatter_bulk}
    {show_save_ax}
    Returns
    -------
    If `show==False` a :class:`~matplotlib.axes.Axes` or a list of it.
    Examples
    --------
    This function behaves very similarly to other embedding plots like
    :func:`~scanpy.pl.umap`
    >>> adata = sc.datasets.visium_sge("Targeted_Visium_Human_Glioblastoma_Pan_Cancer")
    >>> sc.pp.calculate_qc_metrics(adata, inplace=True)
    >>> sc.pl.spatial(adata, color="log1p_n_genes_by_counts")
    See Also
    --------
    :func:`scanpy.datasets.visium_sge`
        Example visium data.
    :tutorial:`spatial/basic-analysis`
        Tutorial on spatial analysis.
    """
    # get default image params if available
    library_id, spatial_data = _check_spatial_data(adata.uns, library_id)
    img, img_key = _check_img(spatial_data, img, img_key, bw=bw)
    spot_size = _check_spot_size(spatial_data, spot_size)
    scale_factor = _check_scale_factor(
        spatial_data, img_key=img_key, scale_factor=scale_factor
    )
    crop_coord = _check_crop_coord(crop_coord, scale_factor)
    na_color = _check_na_color(na_color, img=img)

    if bw:
        cmap_img = "gray"
    else:
        cmap_img = None

    if scale_factor is not None:
        circle_radius = size * scale_factor * spot_size * 0.5
    else:
        circle_radius = spot_size * 0.5

    if figsize is None:
        figsize = (rcParams["figure.figsize"][0], rcParams["figure.figsize"][1])
    topic_names = topic_prop.columns
    topic_prop = topic_prop.to_numpy()

    n_cells = topic_prop.shape[0]
    axs = None
    spatial_coords = adata.obsm[basis]
    n = 0
    colors = plt.get_cmap(cmap)
    colors = colors.colors

    for i in range(n_cells):
        n = n + 1
        if (n % 20000) == 0:
            print(f"{i} number of cells done")
        axs = draw_pie(
            topic_prop[
                i,
            ],
            xpos=spatial_coords[i, 0] * scale_factor,
            ypos=spatial_coords[i, 1] * scale_factor,
            figsize=figsize,
            size=circle_radius,
            ax=axs,
            colors=colors,
        )

    if legend:
        legend_elements = []
        for i in range(len(topic_names)):
            legend_elements.append(
                Patch(facecolor=colors[i], edgecolor="w", label=topic_names[i])
            )

        axs.legend(
            handles=legend_elements,
            bbox_to_anchor=(1, 0.5),
            loc="center left",
            labelspacing=0.5,
            borderpad=0.5,
            frameon=frameon,
        )
    if not isinstance(axs, list):
        axs = [axs]
    for ax in axs:
        cur_coords = np.concatenate([ax.get_xlim(), ax.get_ylim()])
        if img is not None:
            ax.imshow(img, cmap=cmap_img, alpha=alpha_img)
        else:
            ax.set_aspect("equal")
            ax.invert_yaxis()
        if crop_coord is not None:
            ax.set_xlim(crop_coord[0], crop_coord[1])
            ax.set_ylim(crop_coord[3], crop_coord[2])
        else:
            ax.set_xlim(cur_coords[0], cur_coords[1])
            ax.set_ylim(cur_coords[3], cur_coords[2])

    ax.set_xlabel(basis + "1")
    ax.set_ylabel(basis + "2")

    ax.set_xticks([])
    ax.set_yticks([])

    if not frameon:
        ax.axis("off")

    if title is not None:
        ax.set_title(title)

    if show is False or return_fig is True:
        return axs


def get_rgb_function(cmap, min_value, max_value):
    r"""Generate a function to map continous values to RGB values using colormap
    between min_value & max_value."""

    if min_value > max_value:
        raise ValueError("Max_value should be greater or than min_value.")

        # if min_value == max_value:
        #     warnings.warn(
        #         "Max_color is equal to min_color. It might be because of the data or
        #  bad
        #         parameter choice. "
        #         "If you are using plot_contours function try increasing
        # max_color_quantile
        #         parameter and"
        #         "removing cell types with all zero values."
        #     )

        def func_equal(x):
            factor = 0 if max_value == 0 else 0.5
            return cmap(np.ones_like(x) * factor)

        return func_equal

    def func(x):
        return cmap(
            (np.clip(x, min_value, max_value) - min_value) / (max_value - min_value)
        )

    return func


def rgb_to_ryb(rgb):
    """
    Converts colours from RGB colorspace to RYB

    Parameters
    ----------

    rgb
        numpy array Nx3

    Returns
    -------
    Numpy array Nx3
    """
    rgb = np.array(rgb)
    if len(rgb.shape) == 1:
        rgb = rgb[np.newaxis, :]

    white = rgb.min(axis=1)
    black = (1 - rgb).min(axis=1)
    rgb = rgb - white[:, np.newaxis]

    yellow = rgb[:, :2].min(axis=1)
    ryb = np.zeros_like(rgb)
    ryb[:, 0] = rgb[:, 0] - yellow
    ryb[:, 1] = (yellow + rgb[:, 1]) / 2
    ryb[:, 2] = (rgb[:, 2] + rgb[:, 1] - yellow) / 2

    mask = ~(ryb == 0).all(axis=1)
    if mask.any():
        norm = ryb[mask].max(axis=1) / rgb[mask].max(axis=1)
        ryb[mask] = ryb[mask] / norm[:, np.newaxis]

    return ryb + black[:, np.newaxis]


def ryb_to_rgb(ryb):
    """
    Converts colours from RYB colorspace to RGB

    Parameters
    ----------

    ryb
        numpy array Nx3

    Returns
    -------
    Numpy array Nx3
    """
    ryb = np.array(ryb)
    if len(ryb.shape) == 1:
        ryb = ryb[np.newaxis, :]

    black = ryb.min(axis=1)
    white = (1 - ryb).min(axis=1)
    ryb = ryb - black[:, np.newaxis]

    green = ryb[:, 1:].min(axis=1)
    rgb = np.zeros_like(ryb)
    rgb[:, 0] = ryb[:, 0] + ryb[:, 1] - green
    rgb[:, 1] = green + ryb[:, 1]
    rgb[:, 2] = (ryb[:, 2] - green) * 2

    mask = ~(ryb == 0).all(axis=1)
    if mask.any():
        norm = rgb[mask].max(axis=1) / ryb[mask].max(axis=1)
        rgb[mask] = rgb[mask] / norm[:, np.newaxis]

    return rgb + white[:, np.newaxis]


def plot_spatial_general(
    value_df,
    coords,
    labels,
    text=None,
    circle_radius=None,
    display_zeros=False,
    figsize=(10, 10),
    alpha_scaling=1.0,
    max_col=(np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf),
    max_color_quantile=0.98,
    show_img=True,
    img=None,
    img_alpha=1.0,
    adjust_text=False,
    plt_axis="off",
    axis_y_flipped=False,
    x_y_labels=("", ""),
    crop_x=None,
    crop_y=None,
    text_box_alpha=0.9,
    reorder_cmap=range(7),
    style="fast",
    colorbar_position="right",
    colorbar_label_kw={},
    colorbar_shape={},
    colorbar_tick_size=12,
    colorbar_grid=None,
    image_cmap="Greys_r",
    white_spacing=20,
):
    r"""Plot spatial abundance of cell types (regulatory programmes) with colour
    gradient and interpolation.

      This method supports only 7 cell types with these colours (in order, which can be
      changed using reorder_cmap).
      'yellow' 'orange' 'blue' 'green' 'purple' 'grey' 'white'

    :param value_df: pd.DataFrame - with cell abundance or other features (only 7
    allowed, columns) across locations (rows)
    :param coords: np.ndarray - x and y coordinates (in columns) to be used for
    ploting spots
    :param text: pd.DataFrame - with x, y coordinates, text to be printed
    :param circle_diameter: diameter of circles
    :param labels: list of strings, labels of cell types
    :param alpha_scaling: adjust color alpha
    :param max_col: crops the colorscale maximum value for each column in value_df.
    :param max_color_quantile: crops the colorscale at x quantile of the data.
    :param show_img: show image?
    :param img: numpy array representing a tissue image.
        If not provided a black background image is used.
    :param img_alpha: transparency of the image
    :param lim: x and y max limits on the plot. Minimum is always set to 0, if `lim` is
    None maximum
        is set to image height and width. If 'no_limit' then no limit is set.
    :param adjust_text: move text label to prevent overlap
    :param plt_axis: show axes?
    :param axis_y_flipped: flip y axis to match coordinates of the plotted image
    :param reorder_cmap: reorder colors to make sure you get the right color for each
    category

    :param style: plot style (matplolib.style.context):
        'fast' - white background & dark text;
        'dark_background' - black background & white text;

    :param colorbar_position: 'bottom', 'right' or None
    :param colorbar_label_kw: dict that will be forwarded to ax.set_label()
    :param colorbar_shape: dict {'vertical_gaps': 1.5, 'horizontal_gaps': 1.5,
                                    'width': 0.2, 'height': 0.2}, not obligatory to
                                    contain all params
    :param colorbar_tick_size: colorbar ticks label size
    :param colorbar_grid: tuple of colorbar grid (rows, columns)
    :param image_cmap: matplotlib colormap for grayscale image
    :param white_spacing: percent of colorbars to be hidden

    """

    if value_df.shape[1] > 7:
        raise ValueError(
            "Maximum of 7 cell types / factors can be plotted at the moment"
        )

    def create_colormap(R, G, B):
        spacing = int(white_spacing * 2.55)

        N = 255
        M = 3

        alphas = np.concatenate(
            [[0] * spacing * M, np.linspace(0, 1.0, (N - spacing) * M)]
        )

        vals = np.ones((N * M, 4))
        #         vals[:, 0] = np.linspace(1, R / 255, N * M)
        #         vals[:, 1] = np.linspace(1, G / 255, N * M)
        #         vals[:, 2] = np.linspace(1, B / 255, N * M)
        for i, color in enumerate([R, G, B]):
            vals[:, i] = color / 255
        vals[:, 3] = alphas

        return ListedColormap(vals)

    # Create linearly scaled colormaps
    YellowCM = create_colormap(
        240, 228, 66
    )  # #F0E442 ['#F0E442', '#D55E00', '#56B4E9',
    # '#009E73', '#5A14A5', '#C8C8C8', '#323232']
    RedCM = create_colormap(213, 94, 0)  # #D55E00
    BlueCM = create_colormap(86, 180, 233)  # #56B4E9
    GreenCM = create_colormap(0, 158, 115)  # #009E73
    PinkCM = create_colormap(255, 105, 180)  # #C8C8C8
    WhiteCM = create_colormap(50, 50, 50)  # #323232
    PurpleCM = create_colormap(90, 20, 165)  # #5A14A5
    # LightGreyCM = create_colormap(240, 240, 240)  # Very Light Grey: #F0F0F0

    cmaps = [YellowCM, RedCM, BlueCM, GreenCM, PurpleCM, PinkCM, WhiteCM]

    cmaps = [cmaps[i] for i in reorder_cmap]

    with mpl.style.context(style):
        fig = plt.figure(figsize=figsize)
        if colorbar_position == "right":
            if colorbar_grid is None:
                colorbar_grid = (len(labels), 1)

            shape = {
                "vertical_gaps": 1.5,
                "horizontal_gaps": 0,
                "width": 0.15,
                "height": 0.2,
            }
            shape = {**shape, **colorbar_shape}

            gs = GridSpec(
                nrows=colorbar_grid[0] + 2,
                ncols=colorbar_grid[1] + 1,
                width_ratios=[1, *[shape["width"]] * colorbar_grid[1]],
                height_ratios=[1, *[shape["height"]] * colorbar_grid[0], 1],
                hspace=shape["vertical_gaps"],
                wspace=shape["horizontal_gaps"],
            )
            ax = fig.add_subplot(gs[:, 0], aspect="equal", rasterized=True)

        if colorbar_position == "bottom":
            if colorbar_grid is None:
                if len(labels) <= 3:
                    colorbar_grid = (1, len(labels))
                else:
                    n_rows = round(len(labels) / 3 + 0.5 - 1e-9)
                    colorbar_grid = (n_rows, 3)

            shape = {
                "vertical_gaps": 0.3,
                "horizontal_gaps": 0.6,
                "width": 0.2,
                "height": 0.035,
            }
            shape = {**shape, **colorbar_shape}

            gs = GridSpec(
                nrows=colorbar_grid[0] + 1,
                ncols=colorbar_grid[1] + 2,
                width_ratios=[0.3, *[shape["width"]] * colorbar_grid[1], 0.3],
                height_ratios=[1, *[shape["height"]] * colorbar_grid[0]],
                hspace=shape["vertical_gaps"],
                wspace=shape["horizontal_gaps"],
            )

            ax = fig.add_subplot(gs[0, :], aspect="equal", rasterized=True)

        if colorbar_position is None:
            ax = fig.add_subplot(aspect="equal", rasterized=True)

        if colorbar_position is not None:
            cbar_axes = []
            for row in range(1, colorbar_grid[0] + 1):
                for column in range(1, colorbar_grid[1] + 1):
                    cbar_axes.append(fig.add_subplot(gs[row, column]))

            n_excess = colorbar_grid[0] * colorbar_grid[1] - len(labels)
            if n_excess > 0:
                for i in range(1, n_excess + 1):
                    cbar_axes[-i].set_visible(False)

        ax.set_xlabel(x_y_labels[0])
        ax.set_ylabel(x_y_labels[1])

        if img is not None and show_img:
            ax.imshow(img, alpha=img_alpha, cmap=image_cmap)

        # crop images in needed
        if crop_x is not None:
            ax.set_xlim(crop_x[0], crop_x[1])
        if crop_y is not None:
            ax.set_ylim(crop_y[0], crop_y[1])

        if axis_y_flipped:
            ax.invert_yaxis()

        if plt_axis == "off":
            for spine in ax.spines.values():
                spine.set_visible(False)
            ax.tick_params(bottom=False, labelbottom=False, left=False, labelleft=False)

        counts = value_df.values.copy()

        # plot spots as circles
        c_ord = list(np.arange(0, counts.shape[1]))
        colors = np.zeros((*counts.shape, 4))
        weights = np.zeros(counts.shape)

        for c in c_ord:
            min_color_intensity = counts[:, c].min()
            max_color_intensity = np.min(
                [np.quantile(counts[:, c], max_color_quantile), max_col[c]]
            )

            rgb_function = get_rgb_function(
                cmap=cmaps[c],
                min_value=min_color_intensity,
                max_value=max_color_intensity,
            )

            color = rgb_function(counts[:, c])
            color[:, 3] = color[:, 3] * alpha_scaling

            norm = mpl.colors.Normalize(
                vmin=min_color_intensity, vmax=max_color_intensity
            )

            if colorbar_position is not None:
                cbar_ticks = [
                    min_color_intensity,
                    np.mean([min_color_intensity, max_color_intensity]),
                    max_color_intensity,
                ]
                cbar_ticks = np.array(cbar_ticks)

                if max_color_intensity > 13:
                    cbar_ticks = cbar_ticks.astype(np.int32)
                else:
                    cbar_ticks = cbar_ticks.round(2)

                cbar = fig.colorbar(
                    mpl.cm.ScalarMappable(norm=norm, cmap=cmaps[c]),
                    cax=cbar_axes[c],
                    orientation="horizontal",
                    extend="both",
                    ticks=cbar_ticks,
                )

                cbar.ax.tick_params(labelsize=colorbar_tick_size)
                max_color = rgb_function(max_color_intensity / 1.5)
                cbar.ax.set_title(
                    labels[c],
                    **{
                        **{"size": 20, "color": max_color, "alpha": 1},
                        **colorbar_label_kw,
                    },
                )

            colors[:, c] = color
            weights[:, c] = np.clip(counts[:, c] / (max_color_intensity + 1e-10), 0, 1)
            weights[:, c][counts[:, c] < min_color_intensity] = 0

        colors_ryb = np.zeros((*weights.shape, 3))

        for i in range(colors.shape[0]):
            colors_ryb[i] = rgb_to_ryb(colors[i, :, :3])

        def kernel(w):
            return w**2

        kernel_weights = kernel(weights[:, :, np.newaxis])
        weighted_colors_ryb = (colors_ryb * kernel_weights).sum(
            axis=1
        ) / kernel_weights.sum(axis=1)
        weighted_colors = np.zeros((weights.shape[0], 4))
        weighted_colors[:, :3] = ryb_to_rgb(weighted_colors_ryb)
        weighted_colors[:, 3] = colors[:, :, 3].max(axis=1)

        if display_zeros:
            weighted_colors[weighted_colors[:, 3] == 0] = [
                210 / 255,
                210 / 255,
                210 / 255,
                1,
            ]

        ax.scatter(
            x=coords[:, 0], y=coords[:, 1], c=weighted_colors, s=circle_radius**2
        )

        # size in circles is radius
        # add text
        if text is not None:
            bbox_props = dict(boxstyle="round", ec="0.5", alpha=text_box_alpha, fc="w")
            texts = []
            for x, y, s in zip(
                np.array(text.iloc[:, 0].values).flatten(),
                np.array(text.iloc[:, 1].values).flatten(),
                text.iloc[:, 2].tolist(),
            ):
                texts.append(
                    ax.text(x, y, s, ha="center", va="bottom", bbox=bbox_props)
                )

            if adjust_text:
                from adjustText import adjust_text

                adjust_text(texts, arrowprops=dict(arrowstyle="->", color="w", lw=0.5))

    plt.grid(False)
    return fig


def plot_spatial(
    adata,
    topic_prop,
    bw=False,
    img=None,
    basis="spatial",
    library_id=_empty,
    crop_coord=None,
    img_key=_empty,
    spot_size=None,
    na_color=None,
    scale_factor=None,
    scale_default=0.5,
    show_img=True,
    display_zeros=False,
    figsize=(10, 10),
    **kwargs,
):
    """_summary_

    Args:
        adata (_type_): Adata object with spatial coordinates in adata.obsm['spatial']
        topic_prop (_type_): Topic proportion obtained from STAMP.
        bw (bool, optional): bw(currently depreciated). Defaults to False.
        img (_type_, optional): . Defaults to None.
        basis (str, optional): _description_. Defaults to "spatial".
        library_id (_type_, optional): _description_. Defaults to _empty.
        crop_coord (_type_, optional): _description_. Defaults to None.
        img_key (_type_, optional): _description_. Defaults to _empty.
        spot_size (_type_, optional): _description_. Defaults to None.
        na_color (_type_, optional): _description_. Defaults to None.
        scale_factor (_type_, optional): _description_. Defaults to None.
        scale_default (float, optional): _description_. Defaults to 0.5.
        show_img (bool, optional): Whether to display spatial image. Sets to false
        automatically when displaying umap. Defaults to True.
        display_zeros (bool, optional): Whether to display cells that have low
        values to grey colour. Defaults to False.
        figsize (tuple, optional): Figsize of image. Defaults to (10, 10).

    Returns:
        _type_: Function taken from cell2location at
        https://cell2location.readthedocs.io/en/latest/_modules/cell2location/plt/plot_spatial.html#plot_spatial.
        Able to plot both on spatial and umap coordinates. Still very raw.
    """
    # get default image params if available
    library_id, spatial_data = _check_spatial_data(adata.uns, library_id)
    img, img_key = _check_img(spatial_data, img, img_key, bw=bw)
    spot_size = _check_spot_size(spatial_data, spot_size)
    scale_factor = _check_scale_factor(
        spatial_data, img_key=img_key, scale_factor=scale_factor
    )
    crop_coord = _check_crop_coord(crop_coord, scale_factor)
    na_color = _check_na_color(na_color, img=img)

    if scale_factor is not None:
        circle_radius = scale_factor * spot_size * 0.5 * scale_default
    else:
        circle_radius = spot_size * 0.5

    if show_img is True:
        kwargs["show_img"] = True
        kwargs["img"] = img

    kwargs["coords"] = adata.obsm[basis] * scale_factor

    fig = plot_spatial_general(
        value_df=topic_prop,
        labels=topic_prop.columns,
        circle_radius=circle_radius,
        figsize=figsize,
        display_zeros=display_zeros,
        **kwargs,
    )  # cell abundance values

    return fig



import scanpy as sc
import numpy as np
from .utils import check_layer


def filter_genes(adata, min_cutoff=0.01, max_cutoff=1, expression_cutoff_99q=0):
    """Similar function to sc.pp.filter_genes but uses percentage instead of counts.
    Args:
        adata (_type_): Anndata
        min_cutoff (float, optional): Minimum percentage of counts required for a
            gene to pass filtering. Defaults to 0.01.
        max_cutoff (int, optional): Maximum percentage of counts required for a
            gene to pass filtering. Defaults to 0.01.. Defaults to 1.
        expression_cutoff_99q (int, optional): Minimum expression level of gene at
            99th percentile. Defaults to 0.
    """
    n_obs = adata.shape[0]
    min_cells = round(n_obs * min_cutoff)
    max_cells = round(n_obs * max_cutoff)

    sc.pp.filter_genes(adata, min_cells=min_cells)
    sc.pp.filter_genes(adata, max_cells=max_cells)

    data = check_layer(adata, layer=None)

    pass_cutoff = np.quantile(data, q=0.99, axis=0) > expression_cutoff_99q
    genes = adata.var_names[pass_cutoff]
    adata._inplace_subset_var(genes)



import copy

import torch
from torch import Tensor

# from torch_sparse import SparseTensor


class RandomIndexSampler(torch.utils.data.Sampler):
    def __init__(self, num_nodes: int, batch_size: int, shuffle: bool = False):
        self.N = num_nodes
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.n_ids = self.get_node_indices()

    def get_node_indices(self):

        n_ids = torch.randperm(self.N)
        n_ids = list(torch.split(n_ids, self.batch_size))

        return n_ids

    def __iter__(self):
        if self.shuffle:
            self.n_ids = self.get_node_indices()
        return iter(self.n_ids)

    def __len__(self):
        return self.batch_size


class RandomNodeSampler(torch.utils.data.DataLoader):
    r"""A data loader that randomly samples nodes within a graph and returns
    their induced subgraph.

    .. note::

        For an example of using
        :class:`~torch_geometric.loader.RandomNodeSampler`, see
        `examples/ogbn_proteins_deepgcn.py
        <https://github.com/pyg-team/pytorch_geometric/blob/master/examples/
        ogbn_proteins_deepgcn.py>`_.

    Args:
        data (torch_geometric.data.Data): The graph data object.
        batch_size (int): The number of partitions.
        shuffle (bool, optional): If set to :obj:`True`, the data is reshuffled
            at every epoch (default: :obj:`False`).
        **kwargs (optional): Additional arguments of
            :class:`torch.utils.data.DataLoader`, such as :obj:`num_workers`.
    """

    def __init__(self, data, batch_size: int, shuffle: bool = False, **kwargs):
        # assert data.edge_index is not None

        self.N = data.num_nodes
        self.E = data.num_edges

        self.data = copy.copy(data)
        self.data.edge_index = None

        super().__init__(
            self,
            batch_size=1,
            sampler=RandomIndexSampler(self.N, batch_size, shuffle),
            collate_fn=self.__collate__,
            **kwargs,
        )

    def __getitem__(self, idx):
        return idx

    def __collate__(self, node_idx):
        node_idx = node_idx[0]

        data = self.data.__class__()
        data.num_nodes = node_idx.size(0)

        for key, item in self.data:
            if key in ["num_nodes"]:
                continue
            if isinstance(item, Tensor) and item.size(0) == self.N:
                data[key] = item[node_idx]
            else:
                data[key] = item

        return data



import gseapy as gp
import discotoolkit as dt
from joblib import Parallel, delayed


def get_enrichr_geneset(organism="Human"):
    avail_organisms = ["Human", "Mouse", "Yeast", "Fly", "Fish", "Worm"]
    if organism not in avail_organisms:
        raise ValueError(f"available organism are {avail_organisms}")
    return gp.get_library_name(organism=organism)


def get_topic_enrichr(beta, geneset, topics="all", organism="human", topn_genes=20):
    enrichr = {}

    if topics == "all":
        topics = beta.columns
    elif not isinstance(topics, list):
        topics = [topics]
    for topic in topics:
        if topic not in beta.columns:
            raise KeyError(f"{topic} not Found")

    for topic in topics:
        topic_genes = beta.nlargest(topn_genes, topic).index.values.tolist()
        topic_genes = [topic_gene.upper() for topic_gene in topic_genes]
        topic_enrichr = gp.enrichr(
            gene_list=topic_genes, gene_sets=geneset, organism=organism, outdir=None
        )
        if topic not in enrichr:
            enrichr[topic] = topic_enrichr.results
    return enrichr


def get_topic_ora(beta, geneset, topics="all", topn_genes=20, n_jobs=20):

    if topics == "all":
        topics = beta.columns
    elif not isinstance(topics, list):
        topics = [topics]
    for topic in topics:
        if topic not in beta.columns:
            raise KeyError(f"{topic} not Found")

    oras = {}
    topic_genes = []
    for topic in topics:
        topic_genes.append(beta.nlargest(topn_genes, topic).index.values.tolist())

    ora = Parallel(n_jobs=n_jobs)(
        delayed(gp.enrich)(topic_gene, geneset) for topic_gene in topic_genes
    )
    i = 0
    for topic in topics:
        oras[topic] = ora[i].res2d.sort_values("P-value", inplace=False)
        i = i + 1

    return oras

    # for topic in topics:
    #     try:
    #         topic_genes = beta.nlargest(topn_genes, topic).index.values.tolist()
    #         # background = set(list(set(list(beta.index)) - set(topic_genes)))
    #         topic_ora = gp.enrich(
    #             gene_list=topic_genes, gene_sets=geneset, outdir=None, verbose=True
    #         )
    #         if topic not in ora:
    #             ora[topic] = topic_ora.results.sort_values("Adjusted P-value")
    #     except:
    #         ora[topic] = None

    return ora


def get_topic_disco(beta, topics="all", reference=None, topn_genes=20, ncores=20):
    if topics == "all":
        topics = beta.columns.tolist()
    elif not isinstance(topics, list):
        topics = [topics]
    for topic in topics:
        if topic not in beta.columns:
            raise KeyError(f"{topic} not Found")

    def process_topic(beta, topic, reference):
        df = beta[[topic]]
        df.columns = ["fc"]
        df["gene"] = df.index
        df = df.loc[:, ["gene", "fc"]]
        df = df.sort_values("fc", ascending=False)
        df = df.iloc[:topn_genes, :]
        df = df[["gene"]]
        return dt.CELLiD_enrichment(df, ncores=1, reference=reference)

    disco = Parallel(n_jobs=ncores)(
        delayed(process_topic)(beta, topic, reference) for topic in topics
    )

    discos = {}
    for i in range(len(topics)):
        discos[topics[i]] = disco[i]

    return discos


def get_topic_gsea(
    beta,
    geneset,
    topics="all",
    geneset_size=[5, 500],
    n_jobs=1,
    permutations=1000,
):

    if topics == "all":
        topics = beta.columns
    elif not isinstance(topics, list):
        topics = [topics]
    for topic in topics:
        if topic not in beta.columns:
            raise KeyError(f"{topic} not found")

    gsea = {}
    # library = blitz.enrichr.get_library(genesets)

    def process_topic(beta, topic, geneset):
        # Cut off sparse entries
        rank = beta.loc[:, topic]
        rank = rank.sort_values(ascending=False)
        # rank = rank.reset_index()
        results = gp.prerank(
            rank,
            geneset,
            permutations=permutations,
            min_size=geneset_size[0],
            max_size=geneset_size[1],
            threads=1,
            verbose=True,
        )
        return results.res2d

    gsea = Parallel(n_jobs=n_jobs)(
        delayed(process_topic)(beta, topic, geneset) for topic in topics
    )
    # results["Name"] = geneset

    # if topic not in gsea.keys():
    #     gsea[topic] = results

    gseas = {}
    for i in range(len(topics)):
        gseas[topics[i]] = gsea[i]

    return gseas


# def get_niches(adata, topic_prop, resolution=0.5):

#     spatial_connectivities = adata.obsp["spatial_connectivities"]
#     spatial_topic_prop = spatial_connectivities @ topic_prop
#     adata.obsm["X_spatial_latent"] = spatial_topic_prop
#     # sc.pp.neighbors(adata, use_rep = "X_spatial_latent", key_added = )
#     # sc.tl.leiden(adata)

#     return spatial_topic_prop


# def get_topic_msigdb(
#     beta, collection=["hallmark"], test="gsea", genes_filter=[-0.1, 0.1]
# ):
#     beta = beta.transpose()

#     msigdb = dc.get_resource("MSigDB")
#     print("Available options for collection are:", msigdb.collection.unique())
#     msigdb = msigdb[msigdb["collection"].isin(collection)]
#     msigdb = msigdb[~msigdb.duplicated(["geneset", "genesymbol"])]
#     return dc.run_ora(
#         mat=beta, net=msigdb, source="geneset", target="genesymbol", verbose=True
#     )



import numpy as np
import torch
from scipy.sparse import issparse
from torch_geometric.utils import scatter, to_torch_csc_tensor


def check_layer(adata, layer):
    if layer is None:
        if issparse(adata.X):
            data = adata.X.toarray()
        else:
            data = adata.X
    else:
        if issparse(adata.layers[layer]):
            data = adata.layers[layer].toarray()
        else:
            data = adata.layers[layer]
    return data.astype("float32")


def get_init_bg(data):
    # Compute the log background frequency of all words
    # sums = np.sum(data, axis=0)+1
    data = data.copy()
    data = data / data.sum(axis=1, keepdims=True)
    means = np.mean(data, axis=0)
    print("Computing background frequencies")
    return np.log(means + 1e-8)


def precompute_SGC(data, n_layers, add_self_loops=True):

    if n_layers >= 1:
        row, col = data.edge_index
        N = data.num_nodes

        edge_weight = data.edge_weight
        if edge_weight is None:
            edge_weight = torch.ones(data.num_edges, device=row.device)

        deg = scatter(edge_weight, col, dim_size=N, reduce="sum")
        deg_inv_sqrt = deg.pow_(-0.5)
        deg_inv_sqrt.masked_fill_(deg_inv_sqrt == float("inf"), 0)
        edge_weight = deg_inv_sqrt[row] * edge_weight * deg_inv_sqrt[col]
        adj = to_torch_csc_tensor(data.edge_index, edge_weight, size=(N, N))
        adj_t = adj.t()

    assert data.x is not None
    xs = data.x
    xs = torch.log(xs + 1)
    sgc = [xs]
    for i in range(n_layers):
        # xs = matmul(edge_index, xs, reduce="add")
        xs = adj_t @ sgc[-1]
        sgc.append(xs)

    # data["x"] = data.x
    data["sgc_x"] = torch.cat(sgc, dim=1)
    return data



"""Main module."""

import numpy as np
import pandas as pd
import pyro
import pyro.distributions as dist
from pyro.infer import (
    SVI,
    TraceMeanField_ELBO,
)
from pyro.optim import AdamW
import scipy
import torch

# from .data import choose_dataloader
import torch.nn.functional as F
from torch.utils.data import DataLoader
from torch_geometric.data import Data
from torch_geometric.utils import from_scipy_sparse_matrix
from torch_sparse import SparseTensor
from torchinfo import summary
from tqdm import tqdm

# from torch_geometric.loader import RandomNodeSampler
from .data import RandomNodeSampler
from .metrics import get_metrics
from .model import spatialLDAModel
from .utils import (
    check_layer,
    get_init_bg,
    precompute_SGC,
)


class STAMP:
    def __init__(
        self,
        adata,
        n_topics=20,
        n_layers=1,
        hidden_size=50,
        layer=None,
        dropout=0.0,
        categorical_covariate_keys=None,
        continous_covariate_keys=None,
        verbose=True,
        batch_size=1024,
        enc_distribution="mvn",
        beta=1,
    ):
        """_summary_

        Args:
            adata (_type_): AnnData  object
            n_topics (int, optional): Number of topics to model. Defaults to 10.
            n_layers (int, optional): _description_. Defaults to 1.
            hidden_size (int, optional):  Number of nodes in the hidden layer of the
                encoder. Defaults to 50.
            layer (_type_, optional): Layer where the counts data are stored. Defaults
                to None.
            dropout (float, optional): Dropout used for the encoder. Defaults to 0.2.
            batch_key (_type_, optional): Location of batch in obs where the batch
                information is stored. Defaults to None.
            verbose (bool, optional): Print out information on the model. Defaults to
                True.
            batch_size (int, optional): Batch size. Defaults to 1024.
            enc_distribution (str, optional): Encoder distribution. Choices are
                multivariate normal. Defaults to "mvn".
            beta (float, optional): Beta as in Beta-VAE. Defaults to 1.
        """
        pyro.clear_param_store()

        self.continous_covariate_keys = continous_covariate_keys
        self.categorical_covariate_keys = categorical_covariate_keys
        self.hidden_size = hidden_size
        self.n_topics = n_topics
        self.adata = adata
        self.n_cells = adata.shape[0]
        self.n_genes = adata.shape[1]
        self.hidden_size = hidden_size
        self.dropout = dropout
        self.n_layers = n_layers
        self.layer = layer
        self.batch_size = batch_size
        self.enc_distribution = enc_distribution
        self.beta = beta

        bg = get_init_bg(check_layer(adata, layer))
        self.bg_init = torch.from_numpy(bg)

        self.data = self.setup(
            adata, layer, categorical_covariate_keys, continous_covariate_keys
        )

        self.model = spatialLDAModel(
            self.n_genes,
            self.hidden_size,
            self.n_topics,
            self.dropout,
            self.bg_init,
            self.n_layers,
            self.n_batches,
            self.n_cells,
            self.enc_distribution,
            self.beta,
        )

        if verbose:
            print(summary(self.model))

    def setup(self, adata, layer, categorical_covariate_keys, continous_covariate_keys):
        x_numpy = check_layer(adata, layer)
        x = torch.from_numpy(x_numpy)

        if self.n_layers >= 1:
            if "spatial_connectivities" not in adata.obsp.keys():
                raise KeyError("spatial_connectivities not found")

            adj = SparseTensor.from_scipy(
                adata.obsp["spatial_connectivities"]
                + scipy.sparse.identity(n=x_numpy.shape[0])
            )
            adj = adj.t()
            edge_index = from_scipy_sparse_matrix(
                adata.obsp["spatial_connectivities"]
                + scipy.sparse.identity(n=x_numpy.shape[0])
            )[0]
        else:
            adj = None
            edge_index = None

        self.n_batches = 0
        self.one_hot = []
        if categorical_covariate_keys is not None:
            if not isinstance(categorical_covariate_keys, list):
                raise ValueError("categorical_covariate_keys must be a list.")

            for categorical_covariate_key in categorical_covariate_keys:
                self.batch_series = adata.obs[categorical_covariate_key].astype(
                    "category"
                )
                self.n_batches += self.batch_series.nunique()
                batch_factorize, _ = pd.factorize(self.batch_series)
                self.batch_factorize = torch.from_numpy(batch_factorize)
                self.one_hot.append(F.one_hot(self.batch_factorize).float())

        if continous_covariate_keys is not None:
            if not isinstance(continous_covariate_keys, list):
                raise ValueError("continous_covariate_keys must be a list.")

            for continous_covariate_key in continous_covariate_keys:
                self.batch_series = adata.obs[continous_covariate_key].astype("float32")
                self.batch_series = (
                    self.batch_series - self.batch_series.mean()
                ) / self.batch_series.std()
                self.n_batches += 1
                self.batch_factorize = torch.from_numpy(self.batch_series.values)
                self.one_hot.append(self.batch_factorize.float().reshape(-1, 1))

        if self.n_batches == 0:
            self.n_batches += 1
            data = Data(x=x, edge_index=edge_index, adj_t=adj)
        else:
            data = Data(
                x=x,
                edge_index=edge_index,
                adj_t=adj,
                st_batch=torch.cat(self.one_hot, dim=1),
            )

        data = precompute_SGC(data, n_layers=self.n_layers, add_self_loops=True)

        if self.batch_size >= self.n_cells:
            self.dataloader = DataLoader(
                [data],
                batch_size=1,
            )

        else:
            self.dataloader = RandomNodeSampler(
                data, batch_size=self.batch_size, shuffle=True, pin_memory=False
            )
        return data

    def train(
        self,
        max_epochs=2000,
        learning_rate=0.003,
        device="cuda:0",
        weight_decay=0.1,
        early_stop=True,
        patience=20,
    ):
        """Training the data

        Args:
            max_epochs (int, optional): Maximum number of epochs to run.
                Defaults to 2000.
            learning_rate (float, optional): Learning rate of AdamW optimizer.
                Defaults to 0.01.
            device (str, optional): Which device to run model on. Use "cpu"
                to run on cpu and cuda to run on gpu. Defaults to "cuda:0".
            weight_decay (float, optional): Weight decay of AdamW optimizer.
                 Defaults to 0.1.
            early_stop (bool, optional): Whether to early stop when training plateau.
                 Defaults to True.
            patience (int, optional): How many epochs to stop training when
                training plateau. Defaults to 20.
        """

        # adam_args = {"lr":learning_rate, "weight_decay":weight_decay, "clip_norm": 1}
        # optimizer = ClippedAdam(adam_args)
        optimizer = AdamW(
            {"lr": learning_rate, "weight_decay": weight_decay},
            clip_args={"clip_norm": 1},
        )

        self.device = device
        self.model = self.model.to(device)
        pbar = tqdm(range(max_epochs))
        avg_loss = np.Inf
        # tau_prev = 0.5 * torch.ones(self.n_genes, self.n_topics).to(device)
        # tau_prev.require_grad = False
        if early_stop:
            early_stopper = EarlyStopper(patience=patience)
        svi = SVI(
            self.model.model, self.model.guide, optimizer, loss=TraceMeanField_ELBO()
        )

        for epoch in pbar:
            losses = []
            # optimizer.zero_grad()
            for batch_idx, batch in enumerate(self.dataloader):
                batch = batch.to(device)
                if self.n_batches == 1:
                    batch_loss = svi.step(batch.x, batch.sgc_x, None)
                else:
                    batch_loss = svi.step(batch.x, batch.sgc_x, batch.st_batch)

                losses.append(float(batch_loss))

            avg_loss = sum(losses) / self.n_cells
            if np.isnan(avg_loss):
                break
            pbar.set_description(f"Loss:{avg_loss:.3f}")

            if early_stop:
                if early_stopper.early_stop(avg_loss):
                    print("Early Stopping")
                    break

    def get_metrics(self, topk=10, layer=None, TGC=True):
        """Get metrics

        Args:
            topk (int, optional): Number of top genes to use to score the metrics.
                 Defaults to 10.
            layer (_type_, optional): Which layer to use to score the metrics.
                 If none is chosen, use X. Defaults to None.
            TGC (bool, optional): Whether to calculate the topic gene correlation.
                 Defaults to True.

        Returns:
            _type_: _description_
        """
        adata = self.adata
        beta = self.get_feature_by_topic()
        topic_prop = self.get_cell_by_topic()
        metrics = get_metrics(adata, beta, topic_prop, topk=topk, layer=layer, TGC=TGC)

        return metrics

    def get_prior(self, device="cuda:0"):
        self.model.to(device)
        self.data.to(device)
        prior_loc, prior_scale = self.model.get_prior(self.data.st_batch)
        return prior_loc.detach().cpu().numpy(), prior_scale.detach().cpu().numpy()

    def get_dispersion(self, device="cuda:0"):
        model = self.model.model_params()
        # model.eval()
        model.to(device)
        self.model.set_device(device)
        self.model.eval()
        # self.model.to(device)
        self.data.to(device)

        pred = self.model.predictive(num_samples=10)
        if self.n_batches > 1:
            pred = pred(
                self.data.x, self.data.adj_t, self.data.sgc_x, self.data.st_batch
            )
        else:
            pred = pred(self.data.x, self.data.adj_t, self.data.sgc_x)
        # rate = self.model.feature("rate")
        # rate = torch.exp(rate)
        # rate = rate.detach().cpu().numpy()
        rate = pred["disp"].mean(axis=0).detach().cpu().numpy()
        # ads = pred["ads"].mean(axis=0).detach().cpu().numpy()
        df = pd.DataFrame(rate, index=self.adata.var_names, columns=["disp"])
        # df["ads"] = ads
        return df

    def save(self, path):
        self.model.save(path)

    def load(self, path):
        self.model.load(path)
        self.model.eval()

    def get_cell_by_topic(self, device="cpu"):
        """Get latent topics after training.

        Args:
            device (str, optional): What device to use. Defaults to "cpu".

        Returns:
            _type_: A dataframe of cell by topics where each row sum to one.
        """
        model = self.model.model_params()
        model.eval()
        model.to(device)
        self.data.to(device)

        with torch.no_grad():
            # if self.batch_key is None:
            x = self.data.x
            if (self.continous_covariate_keys is None) and (
                self.categorical_covariate_keys is None
            ):
                cell_topic = self.model.guide(x, self.data.sgc_x)
            else:
                cell_topic = self.model.guide(x, self.data.sgc_x, self.data.st_batch)

            cell_topic = cell_topic.detach().cpu().numpy()
            cell_topic = pd.DataFrame(
                cell_topic,
                columns=["Topic" + str(i) for i in range(1, self.n_topics + 1)],
            )
            cell_topic.set_index(self.adata.obs_names, inplace=True)

            return cell_topic

    def get_feature_by_topic(
        self,
        device="cpu",
        num_samples=1000,
        pct=0.5,
        return_softmax=False,
    ):
        """Get the gene modules

        Args:
            device (str, optional): Which device to use. Defaults to "cpu".
            num_samples (int, optional): Number of samples to use for calculation.
              Defaults to 1000.
            pct (float, optional): ??? . Defaults to 0.5.
            return_softmax (bool, optional): ???. Defaults to False.

        Returns:
            _type_: _description_
        """
        # pseudcount of 0.5 aga aga
        # feature_topic = feature_topic.t()
        # Transform to torch log scale
        self.model.to(device)
        self.data.to(device)

        # feature_topic = self.model.feature_by_topic()
        # feature_topic = feature_topic.to(device)
        if pct == 0.5:
            feature_topic = self.model.feature_by_topic(
                return_scale=False, return_softmax=return_softmax
            )
            feature_topic = feature_topic.to(device)
        else:
            feature_topic_loc, feature_topic_scale = self.model.feature_by_topic(
                return_scale=True
            )
            feature_topic = dist.Normal(
                feature_topic_loc, torch.sqrt(torch.exp(feature_topic_scale))
            )
            feature_topic = feature_topic.sample((num_samples,)).kthvalue(
                int(pct * num_samples), dim=0
            )[0]
            feature_topic = feature_topic.to(device)
        # feature_topic = torch.sqrt(torch.exp(feature_topic_scale))
        # if pseudocount > 0:
        #     if self.n_batches > 1:
        #         bg = self.model.get_bias().min(axis=0)[0]
        #     else:
        #         bg = self.model.get_bias()
        #     amt = torch.log(bg.exp() + pseudocount) - bg
        #     feature_topic = feature_topic.t() - amt
        #     feature_topic = feature_topic.t()

        feature_topic = feature_topic.detach().cpu().numpy()

        feature_topic = feature_topic[:, : self.n_topics]
        feature_topic = pd.DataFrame(
            feature_topic,
            columns=["Topic" + str(i) for i in range(1, self.n_topics + 1)],
            index=self.adata.var_names,
        )

        return feature_topic

    def get_background(self):
        background = self.model.get_bias().detach().cpu().numpy()
        background = pd.DataFrame(
            background, index=self.adata.var_names, columns=["gene"]
        )
        return background

    # def plot_qc(self, n_obs=1000, gene=None, device="cuda:0"):
    #     self.model.eval()
    #     self.model.to(device)
    #     self.data.to(device)

    #     if gene is None:
    #         gene = self.adata.var_names
    #     gene_index = np.where(self.adata.var_names == gene)[0]

    #     x_plot = self.data.x.detach().cpu().numpy()
    #     obs = random.sample(range(x_plot.shape[0]), n_obs)
    #     x_plot = x_plot[obs, :]
    #     x_plot = x_plot / x_plot.sum(axis=1, keepdims=True)

    #     if self.batch_key is None:
    #         w = self.get_feature_by_topic(return_softmax=True)
    #         z = self.get_cell_by_topic()
    #         w = w.to_numpy()
    #         z = z.to_numpy()
    #         z = z[obs, :]
    #         mean = z @ w.transpose()
    #         mean = mean
    #     else:
    #         w = self.get_feature_by_topic(return_softmax=True)
    #         z = self.get_cell_by_topic()
    #         w = w.to_numpy()
    #         z = z.to_numpy()
    #         z = z[obs, :]
    #         ys = self.data.st_batch.detach().cpu().numpy()
    #         ys = ys[obs, :]
    #         z = np.hstack([z, ys])
    #         mean = z @ w.transpose()
    #         # if n_obs > x.shape[0]:
    #     #     n_obs = x.shape[0]
    #     x_plot = x_plot[:, gene_index].ravel()
    #     y_plot = mean[:, gene_index].ravel()

    #     plt.hist2d(
    #         x_plot,
    #         y_plot,
    #         bins=100,
    #         norm=matplotlib.colors.LogNorm(),
    #     )


class EarlyStopper:
    def __init__(self, patience=50, min_delta=0):
        self.patience = patience
        self.min_delta = min_delta
        self.counter = 0
        self.min_training_loss = np.inf

    def early_stop(self, training_loss):
        if training_loss < self.min_training_loss:
            self.min_training_loss = training_loss
            self.counter = 0
        elif training_loss > (self.min_training_loss + self.min_delta):
            self.counter += 1
            if self.counter >= self.patience:
                return True
        return False
